#import "WiiJunker.h"
#import <DiscRecordingUI/DiscRecordingUI.h>
#include <sys/types.h>
#include <sys/stat.h>


static WiiJunker *sharedInstance = nil;

@implementation WiiJunker

+ (id)getInstance
{
	return sharedInstance;
}

- (void)awakeFromNib
{
	sharedInstance =self;
	[window center];
	[infoField setHidden:YES];
	[infoField setStringValue:@""];
	[spinner setDisplayedWhenStopped:NO];
	[patch setState:NSOffState];
	[patch setEnabled:NO];
	[burn setEnabled:NO];
	burning = NO;
	[apply setEnabled:NO];
	iso=NULL;
	unsc = nil;
	unscramblingISOPath = nil;
	unscrambling = NO;
	[NSApp setDelegate:self];
}

- (void)dealloc
{
	[self reset];
	[super dealloc];
}

- (void)reset
{
	if (path){
		[path release];
		path=nil;
	}
	
	[pathField setStringValue:@""];
	[gameName setStringValue:@""];
	[gameRegion setStringValue:@""];
	[updateField setStringValue:@""];
	[patch setState:NSOffState];
	[patch setEnabled:NO];
	if (iso){
		fclose(iso);
		iso = NULL;
	}
	
	if(unsc)
		[unsc release];
	unsc = nil;
	
	if (unscramblingISOPath)
		[unscramblingISOPath release];
}


- (IBAction)openISO:(id)sender
{
	
	NSOpenPanel *oPanel = [NSOpenPanel openPanel];
	
    [oPanel setAllowsMultipleSelection:NO];
    [oPanel setCanChooseDirectories:NO];
    [oPanel setCanChooseFiles:YES];
    [oPanel setResolvesAliases:NO];
    
    [oPanel beginSheetForDirectory:nil file:nil types:[NSArray arrayWithObjects:@"iso", @"cdr", @"wod", @"wii", nil]
                    modalForWindow:[NSApp mainWindow]
                     modalDelegate: self
                    didEndSelector: @selector(openPanelDidEnd:returnCode:contextInfo:)
                       contextInfo: nil];
					   
	
}

- (void)openPanelDidEnd:(NSOpenPanel *)sheet returnCode:(int)returnCode contextInfo:(void *) contextInfo
{
	[sheet orderOut:nil];
	
    if (returnCode == NSOKButton){
		[self reset];
        path = [[[sheet filenames] objectAtIndex:0] retain];
		if (path ==nil && [path length] < 1)
		{
			return;
		}
		
		[self checkType];
		
    }
	[NSApp stopModal];
}

- (void)checkType
{
	if ([self isWiiFile])
	{
		[self unscStart];
	} else {
		if (![self validateSize])
		{
			return;
		}
		
		if(!(iso = fopen([path UTF8String], "r+"))) {
			perror([path UTF8String]);
			return;
		}
		
		[apply setEnabled:YES];
	
		[self loadInfo];
	}
}

- (BOOL)isWiiFile
{
	struct stat *st = malloc( sizeof( struct stat ) );	
	if (iso ==NULL)
		stat([path UTF8String], st);
	
	BOOL size_ok =  ((st->st_size == 4736698368LL) ? YES: NO);
	free(st);
	return size_ok;
}

- (BOOL)validateSize
{
	struct stat *st = malloc( sizeof( struct stat ) );	
	if (iso ==NULL)
		stat([path UTF8String], st);
	
	BOOL size_ok =  ((st->st_size == 4699979776LL) || (st->st_size == 4700012544LL) ? YES: NO);
	if (!size_ok){
		if (NSRunAlertPanel(NSLocalizedString(@"ISO size is weird !", @"ISO size is weird !"),
							NSLocalizedString(@"It should be should be 4 699 979 776 or 4 700 012 544 bytes, are you sure you want to continue ?", @"It should be 4 699 979 776 bytes, are you sure you want to continue ?"),
							NSLocalizedString(@"Cancel", @"Cancel"),NSLocalizedString(@"Continue", @"Continue"),nil) != NSAlertDefaultReturn)
		{
			size_ok=YES;
		} 
			
	}
	free(st);
	return size_ok;
}
#pragma mark -
#pragma mark Unscrambling

- (void)unscStart
{
	unscrambling = YES;
	if(unsc)
		[unsc release];
		
	unsc = [[Unscrambler alloc] init];
	
	NSSavePanel *sPanel = [NSSavePanel savePanel];
	[sPanel setExtensionHidden:NO];
    [sPanel setRequiredFileType:@"iso"];

	NSString *filename= [NSString stringWithFormat:@"%@.iso", 
										[[path lastPathComponent] stringByDeletingPathExtension]];
    [sPanel beginSheetForDirectory:nil file:filename 
                    modalForWindow:[NSApp mainWindow]
                     modalDelegate: self
                    didEndSelector: @selector(savePanelDidEnd:returnCode:contextInfo:)
                       contextInfo: nil];
	
}

- (void)savePanelDidEnd:(NSSavePanel *)sheet returnCode:(int)returnCode contextInfo:(void *) contextInfo
{
    [sheet orderOut:nil];
	[NSApp stopModal];
    if (returnCode == NSOKButton){
        unscramblingISOPath = [[sheet filename] retain];
		if (unscramblingISOPath ==nil && [unscramblingISOPath length] < 1)
		{
			[unsc release];
			unscrambling = NO;
			return;
		}
		[self showUnscProgress];
    }
	
}

- (void)unscThread
{
	NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
	int ret=[unsc unscramble:[path UTF8String] toIso:[unscramblingISOPath UTF8String]];
	
	if (ret==0)
	{
		[NSApp endSheet:unscWindow];
		[path release];
		path = nil;
		path = [[NSString stringWithString:unscramblingISOPath] retain];
		[unscramblingISOPath release];
		unscramblingISOPath = nil;
		unscrambling = NO;
		//[unsc release];
		[self checkType];
		
	} 
	
	[pool release];
	
}

- (void)showUnscProgress
{
	
	[NSThread detachNewThreadSelector:@selector(unscThread) toTarget:self withObject:nil];

	NSString *info=[NSString stringWithFormat:NSLocalizedString(@"Processing %@...", @"Processing %@..."), [path lastPathComponent]];
	[unscISOField setStringValue:info];
	[unscProgress setDoubleValue:0.0];
	[unscStatus setStringValue:@"0%"];
	[NSApp beginSheet: unscWindow
            modalForWindow: window
            modalDelegate: self
            didEndSelector: @selector(didEndSheet:returnCode:contextInfo:)
            contextInfo: nil];
}



- (void)updateProgress:(float)value andStatus:(NSString*)status
{
	if (value > -1)
		[unscProgress setDoubleValue:[[NSNumber numberWithFloat:value] intValue]];
	
	NSString *st = [NSString stringWithFormat:@"%.2f%%", (value > -1) ? value : 0];
	if (status!=nil){
		st = [st stringByAppendingFormat:@", %@", status];
	}
	
	[unscStatus setStringValue:st];
}

- (void)unscFinish
{
	[NSApp endSheet:unscWindow];
	unscrambling = NO;
	[unsc release];
	
	
	[unscISOField setStringValue:@""];
	[unscProgress setDoubleValue:0.0f];
	[unscStatus setStringValue:@""];
}

- (void)unscFailedWithError:(NSString *)err
{
	NSRunCriticalAlertPanel(NSLocalizedString(@"Unscrambling failed !", @"Unscrambling failed !"),
							err,nil,nil,nil);
	
	[NSApp endSheet:unscWindow];
	unscrambling = NO;
	[unsc interrupt];
	//[unsc release];
	
	[unscISOField setStringValue:@""];
	[unscProgress setDoubleValue:0.0f];
	[unscStatus setStringValue:@""];
}

- (IBAction)abortUnscrambler:(id)sender
{
	if (NSRunCriticalAlertPanel(NSLocalizedString(@"Abort unscrambling process ?", @"Abort unscrambling process ?"),
							NSLocalizedString(@"Are you sure you want to abort this process ? You cannot burn your game if you do not unscramble it.", @"Are you sure you want to abort this process ? You cannot burn your game if you do not unscramble it."),
							NSLocalizedString(@"Continue", @"Continue"), NSLocalizedString(@"Really Abort", @"Really Abort"),nil) == NSAlertAlternateReturn){
		[NSApp endSheet:unscWindow];
		[unsc interrupt];
		//[unsc release];
		unscrambling = NO;
		[path release];
		path=nil;
		[unscramblingISOPath release];
		unscramblingISOPath = nil;
		
		[unscISOField setStringValue:@""];
		[unscProgress setDoubleValue:0.0f];
		[unscStatus setStringValue:@""];
		[self reset];
	}
}

- (void)didEndSheet:(NSWindow *)sheet returnCode:(int)returnCode contextInfo:(void *)contextInfo
{
    [sheet orderOut:self];
}

#pragma mark -
#pragma mark ISO related stuff


- (void)loadInfo
{
	
	[pathField setStringValue:path];
	[gameName setStringValue:[self getGameName]];
	[gameRegion setStringValue:[self getGameRegion]];
	[updateField setStringValue:[self includesUpdate] ? @"YES" : @"NO"];
	 
	if ([self includesUpdate])
	{
		[patch setState:NSOnState];
		[patch setEnabled:YES];
	} else {
		[patch setState:NSOffState];
		[patch setEnabled:NO];
	}
	[burn setEnabled:YES];
	[apply setEnabled:YES];
}


- (NSString *)getGameName
{
	char buf[256];
	fseek(iso, 0x20, SEEK_SET);
	fread(buf, sizeof(buf), 1, iso);
	rewind(iso);
	NSMutableString *s = [[NSMutableString alloc] init];
	int i = 0;
	
	for(i = 0; i < 256 && buf[i] != 0; i++)
		[s appendFormat:@"%c", (char)buf[i]];
	NSString *result = [NSString stringWithString:s];
	[s release];
	
	return result;
	
}

- (NSString *)getGameRegion
{
	NSString *regions[3]={@"Japan", @"USA", @"Europe"};
	
	int code = [self getRegionCode];
	return regions[code];
}

- (int)getRegionCode
{
	uint32_t b ;
	
	fseek(iso, 0x4E000, SEEK_SET);
	fread(&b, sizeof(uint32_t), 1, iso);
	b = NSSwapBigIntToHost(b);
	
	rewind(iso);
	if(b > 2)
            b = 2;
	return b;
}

- (BOOL)includesUpdate
{
	uint32_t b;
	fseek(iso, 0x58010, SEEK_SET);
	fread(&b, sizeof(uint32_t), 1, iso);
	b=NSSwapBigIntToHost(b);
	rewind(iso);
	
	return (b!=-514882163 ? YES: NO);
	
}

#pragma mark -
#pragma mark Region / Patch ISO stuff

- (void)setRegion:(int) regionCode
{
	if (regionCode == [self getRegionCode])
	{
		if (NSRunAlertPanel(NSLocalizedString(@"Same Region Code Detected", @"Same Region Code Detected"),
							NSLocalizedString(@"Are you sure you want to continue ?", @"Are you sure you want to continue ?"),
							NSLocalizedString(@"Continue", @"Continue"),NSLocalizedString(@"Cancel", @"Cancel"),nil) != NSAlertDefaultReturn) 
			return;
	}
	
	NSString *regionString[3] = {@"00808080808080808080808080808080",		// Japan 
								 @"80068080808080808080808080808080",		// USA
								 @"80808000030304030080808080808080" };		//Europe
	
	/*
	NSString *regionString[3] = {@"00000000000000000000000000000000",		// Japan 
								 @"00000000000000000000000000000000",		// USA
								 @"00000000000000000000000000000000" };		//Europe
	
	
	*/
	
	if(regionCode < 0 || regionCode > 2){
		NSLog(@"*** DANGER *** Wrong Region Code supplied (%d)", regionCode);
		return;
	}
	
	
	regionCode = NSSwapBigIntToHost(regionCode);
	fseek(iso, 0x4E000, SEEK_SET);
	fwrite(&regionCode, sizeof(int), 1, iso);
	rewind(iso);
	
	
	NSString *r = regionString[regionCode];
	char byte[16];
	int j = 0;
	
	for(j=0; j < 16; j++)
	{
		char byte0 = (char)[r characterAtIndex:((j << 1) + 0)];
		char byte1 = (char)[r characterAtIndex:((j << 1) + 1)];
		byte0 -= 48;
		byte1 -= 48;
		byte[j] = (char)(byte0 << 4 | byte1);
	}

	fseek(iso, 0x4E010, SEEK_SET);
	fwrite(byte, 16, 1, iso);
	rewind(iso);
}


#define BUFSIZE 0x40000

- (void)patchUpdate
{
	
	NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
	
	FILE *dump;
	NSString *dumpPath = [[NSBundle mainBundle] pathForResource:@"clean-dump" ofType:nil];
	if(!(dump = fopen([dumpPath UTF8String], "r"))) {
		perror([dumpPath UTF8String]);
		[pool release];
		return;
	}
	
	unsigned char buf[BUFSIZE];
	
	int i;

	fseek(iso, 0x00050000, SEEK_SET);
	for(i=0; i<8; i++) {
		if(fread(buf, BUFSIZE, 1, dump) <= 0 || fwrite(buf, BUFSIZE, 1, iso) <= 0){       
			NSLog(@"Error Patching to ISO file");
			fclose(dump);
			dump =NULL;
			[pool release];
			return;
		}	
	}
	
	if(fread(buf, 0x10000, 1, dump) <= 0 || fwrite(buf, 0x10000, 1, iso) <= 0){       
		NSLog(@"Error Patching to ISO file");
		fclose(dump);
		dump =NULL;
		[pool release];
		return;
	}
	
	memset(buf, 0, BUFSIZE);
	if(fwrite(buf, 0x20000, 1, iso) <= 0) {       
		NSLog(@"Error Patching to ISO file");
		fclose(dump);
		dump =NULL;
		[pool release];
		return;
	}
	for(i=0; i<982; i++) {
		if(fwrite(buf, BUFSIZE, 1, iso) <= 0){       
			NSLog(@"Error Patching to ISO file");
			fclose(dump);
			dump =NULL;
			[pool release];
			return;
		}
	}
	
	rewind(iso);
	
	fclose(dump);
	dump = NULL;
	
	[infoField setHidden:YES];
	[infoField setStringValue:@""];
	[spinner stopAnimation:nil];
	[apply setEnabled:YES];
	[self loadInfo];
	[pool release];
	
}


- (IBAction)applyChanges:(id)sender
{
	int newRegion = [regionPopUp indexOfSelectedItem];
	[self setRegion:newRegion];
	
	[apply setEnabled:NO];
	if ([patch state] == NSOnState)
	{
		[infoField setHidden:NO];
		[infoField setStringValue:@"Patching ISO file ... Please wait"];
		[spinner startAnimation:nil];
		
		[NSThread detachNewThreadSelector:@selector(patchUpdate) toTarget:self withObject:nil];
	}else {
		[self loadInfo];
	}
	
}


#pragma mark -
#pragma mark Burning Stuff

- (IBAction)burn:(id)sender
{
	
	if (path !=nil && [path length] > 1)
	{
		DRBurnSetupPanel*  bsp = [DRBurnSetupPanel setupPanel];
		[bsp setDelegate:self];
		[bsp setCanSelectAppendableMedia:NO];
		[bsp beginSetupSheetForWindow:window modalDelegate:self didEndSelector:@selector(burnSetupEnded:returnCode:contextInfo:) contextInfo:nil];
	}
}

- (BOOL)setupPanel:(DRSetupPanel*)aPanel deviceContainsSuitableMedia:(DRDevice*)device promptString:(NSString**)prompt
{
	NSString*		mediaType;
	// check to see what sort of media there is present in the drive.
	mediaType = [[[device status] objectForKey:DRDeviceMediaInfoKey] objectForKey:DRDeviceMediaTypeKey];
	if ([mediaType isEqualToString:DRDeviceMediaTypeDVDR] == NO 
	&&  [mediaType isEqualToString:DRDeviceMediaTypeDVDPlusR] == NO
	&&  [mediaType isEqualToString:DRDeviceMediaTypeDVDRW] == NO
	&&  [mediaType isEqualToString:DRDeviceMediaTypeDVDPlusRW] == NO)
	{
		*prompt = [NSString stringWithCString:"Please insert a blank DVD."];
		return NO;
	}
	
	
	*prompt = [NSString stringWithCString:"Ready to burn."];

	return YES;

}

- (void)burnSetupEnded:(DRBurnSetupPanel*)panel returnCode:(int)returnCode contextInfo:(void*)contextInfo
{
	[panel orderOut:self];
	
	if (returnCode == NSOKButton)
	{
		DRBurnProgressPanel*  bpp = [DRBurnProgressPanel progressPanel];
		DRBurn *dvd = [panel burnObject];
		[bpp setVerboseProgressStatus:YES];
		[bpp setDescription:[NSString stringWithFormat:@"Burning %@", [gameName stringValue]]];
		[[DRNotificationCenter currentRunLoopCenter] addObserver:self selector:@selector(burnNotification:) name:DRBurnStatusChangedNotification object:dvd];
		[bpp beginProgressSheetForBurn:dvd layout:[DRBurn layoutForImageFile:path] modalForWindow:window];
	}
	
}

- (void)burnNotification:(NSNotification*)notification	
{
	/*
	NSDictionary* status = [notification userInfo];

	if ([[status objectForKey:DRStatusStateKey] isEqualTo:DRStatusStateDone])
		[[NSNotificationCenter defaultCenter] postNotificationName:@"growlFinishedBurning" object:[[nameField stringValue] stringByAppendingString:NSLocalizedString(@" was burned successfully",@"Localized")]];
	
	if ([[status objectForKey:DRStatusStateKey] isEqualTo:DRStatusStateFailed])
		[[NSNotificationCenter defaultCenter] postNotificationName:@"growlFailedBurning" object:[NSLocalizedString(@"Failed to burn ",@"Localized") stringByAppendingString:[nameField stringValue]]];
	*/
	
}

- (void) burnProgressPanelWillBegin:(NSNotification*)aNotification
{
	burning = YES;	// Keep the app from being quit from underneath the burn.
}

- (void) burnProgressPanelDidFinish:(NSNotification*)aNotification
{
	burning = NO;	// OK we can quit now.
}


#pragma mark -
#pragma mark App / Window delegates

- (NSApplicationTerminateReply)applicationShouldTerminate:(NSApplication *)sender
{
	return (burning || unscrambling ? NSTerminateCancel : NSTerminateNow );
}


-(BOOL)windowShouldClose:(id)sender
{
	if ([sender isEqualTo:window]){
		[NSApp hide:nil];
		return NO;
	}
	
	return YES;
}



@end
