/*
unscrambler 0.4: unscramble not standard IVs scrambled DVDs thru 
bruteforce, intended for Gamecube/WII Optical Disks.

Copyright (C) 2006  Victor Muoz (xt5@ingenieria-inversa.cl)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


** Modified main function for included into WiiJunkies - Junglepop 01/04/08

*/

#include <stdio.h>
#include <string.h>

#include "ecma-267.h"
#import "Unscrambler.h"
#import "WiiJunker.h"

#define MAX_SEEDS 4

unsigned char b_in[0x8100];
unsigned char b_out[0x8000];

typedef struct t_seed {
    int seed;
    unsigned char streamcipher[2048];
} t_seed;


t_seed _seeds[(MAX_SEEDS+1)*16];

/* swap the endianess of a 32bit integer */
unsigned int swap32(unsigned int p) {
	return NSSwapBigIntToHost(p);
	
/*#ifdef __BIG_ENDIAN__
	return p;
#endif
#ifdef __LITTLE_ENDIAN__
    return p<<24|((p<<8)&0xFF0000)|((p>>8)&0xFF00)|(p>>24);
#endif*/
}

/* add a seed to the cache */
t_seed *add_seed(t_seed *seeds, unsigned short seed) {
    int i;
        
    //printf("caching seed %04x\n", seed);

    if(seeds->seed==-2) return NULL;

    seeds->seed=seed;
      
    LFSR_init(seeds->seed);

    for(i=0; i<2048; i++) seeds->streamcipher[i]=LFSR_byte();
    
    return seeds;
}

/* test if the current seed is the one used for this sector, the check is done 
comparing the EDC generated, with the one at the bottom of sector */
int test_seed(int j) {
    int i;
    unsigned char buf[2064];
      
    LFSR_init(j);

    memcpy(buf, b_in, 2064);
    for(i=12; i<2060; i++) buf[i]^=LFSR_byte();
    if(edc_calc(0x00000000, buf, 2060) == swap32(*( (unsigned int *) (&b_in[2060]) )) ) {
        return 0;
    }
    return -1;
}

/* unscramble a complete frame, based on the seed already cached */
int unscramble_frame(t_seed *seed, unsigned char *_bin, unsigned char *_bout) {
    unsigned char *bin;
    unsigned char *bout;
    unsigned int edc;
    unsigned int *_4bin;
    unsigned int *_4cipher;
    
    int i,j;
    
    for(j=0; j<16; j++) {
      
        bin=&_bin[0x810*j];
        bout=&_bout[0x800*j];
        
        _4bin=(unsigned int *)&bin[12];
        _4cipher=(unsigned int *)seed->streamcipher;

        for(i=0; i<512; i++) _4bin[i]^=_4cipher[i];
        memcpy(bout, bin+6, 2048); // copy CPR_MAI bytes
        
        edc=edc_calc(0x00000000, bin, 2060);
        if(edc != swap32(*( (unsigned int *) (&bin[2060]) )) ) {
            printf("error: bad edc (%08x) must be %08x\n", edc, swap32(*( (unsigned int *) (&bin[2060]) )));
            return -1;
        }
    }
    
    return 0;
}



@implementation Unscrambler 
- (void)dealloc
{
	if(start){
		[start release];
		start = nil;
	}
	[super dealloc];
}

- (int)unscramble:(const char *)wod toIso:(const char *)iso
{
	int i,j,s;
    int ret;
    interrupt = NO;
    FILE *in, *out;
    WiiJunker *wii = [WiiJunker getInstance];
    [wii updateProgress:0.0 andStatus:@"estimating time remaining..."];
	
    t_seed *seeds;
    t_seed *current_seed;
    
    in=fopen(wod,"rb");
    if(!in) {
        //fprintf(stderr, "can't open %s\n", wod);
		[wii unscFailedWithError:[NSString stringWithFormat:@"Error opening WOD (%s)", wod]];
        return 1;
    }
    out=fopen(iso,"wb");
    if(!out) {
        //fprintf(stderr, "can't open %s\n", argv[2]);
		[wii unscFailedWithError:[NSString stringWithFormat:@"Error opening ISO destination file (%s)", iso]];
        return 2;
    }
    
	
    for(i=0; i<16; i++) {
        for(j=0; j<MAX_SEEDS; j++) {
            _seeds[i*MAX_SEEDS+j].seed=-1;
        }
        _seeds[i*MAX_SEEDS+j].seed=-2;
    }
    
    ret=0;
    
    s=0;
    start=nil;
	NSString *timeLeft=nil;
	byteCopied = 0;
	byteRemaining = 4699979776LL;
	iso_size=4700012544LL; //4699979776LL;
	unsigned long updateDisplay = 0;
	BOOL updateCaching =YES;
    do {
		
        fread(b_in, 1, 0x8100, in);
        
        seeds=&_seeds[((s>>4)&0xF)*MAX_SEEDS];
        while((seeds->seed)>=0) {
            if(!test_seed(seeds->seed)) {
                current_seed=seeds;
                goto seed_found;
            }
            seeds++;
        }
        
        for(j=0; j<0x7FFF; j++) {
            if(!test_seed(j)) {
                current_seed=add_seed(seeds,j);
                //printf("caching at %x\n", s);
                if(current_seed==NULL) {
                    //fprintf(stderr, "no enough cache space for this seed.\n");
					[wii unscFailedWithError:[NSString stringWithFormat:@"Not enough cache space for seed at %x", s]];
                    ret=3;
                    goto finish;
                }
				if(updateCaching)
					[wii updateProgress:-1 andStatus:@"caching seeds..."];
                goto seed_found;
            }
        }
        
		//fprintf(stderr, "no seed found for recording frame %d.\n", s>>4);
        ret=4;
        goto finish;
        
        seed_found:
        
        if(unscramble_frame(current_seed, b_in, b_out)) {
            //fprintf(stderr, "error unscrambling recording frame %d.\n", s>>4);
			[wii unscFailedWithError:[NSString stringWithFormat:@"Error unscrambling recording frame %d.\n", s>>4]];
            ret=5;
            goto finish;
        }
        
        if(fwrite(b_out, 1, 0x8000, out)!=0x8000) {
            //fprintf(stderr, "can't write to the output file, check if there is enough free space.\n");
            [wii unscFailedWithError:[NSString stringWithFormat:@"Can't write to the output file, check if there is enough free space"]];
       
			ret=6;
            goto finish;        
        } else {
			byteCopied+=32768;
			byteRemaining-=32768;
			
			if (!interrupt){
					
				float progress = (float)((float)byteCopied*100)/iso_size;
				//NSLog(@"c: %llu, iso: %llu, p: %.2f", byteCopied, iso_size, progress);
				updateDisplay+=32768;
				if ((updateDisplay > 327680) && (start == nil))
					start=[[NSCalendarDate calendarDate] retain];
					
				if (updateDisplay > 4699978)
				{
					timeLeft = [self timeRemaining];
					updateCaching = NO;
					[wii updateProgress:progress andStatus:timeLeft];
					updateDisplay=0;
				}
			}
			
		}
		
		
		
        s+=16;
    } while(!feof(in) && !interrupt);
    
    //printf("image successfully unscrambled.\n");
    
    finish:
    
    //printf("time elapsed: %.2lf seconds.\n", difftime(time(0), start));
	
    fclose(in);
    fclose(out);
	if (interrupt)	
		ret = 7;
    return ret;
}


- (NSString *)timeRemaining
{
	
	NSCalendarDate *now = [NSCalendarDate calendarDate];
	NSTimeInterval dTime = [now timeIntervalSinceDate:start];
	
	unsigned long timeRemaining = (dTime*byteRemaining)/byteCopied;
	
	NSMutableString *format = [NSMutableString stringWithString:@""];
	int sec = (timeRemaining) % 60;
	int min = (timeRemaining/60) % 60;
	int hours = (timeRemaining/(60*60)) % 24;
	int days = (timeRemaining/(60*60*24)) % 7;
	int weeks = (timeRemaining/(60*60*24*7));
	
	if (weeks !=0)
		[format appendFormat:@"%d weeks ", weeks];
	if (days!=0)
		[format appendFormat:@"%d days ", days];
	if (hours!=0)
		[format appendFormat:@"%d hours ", hours];
	if (min!=0)
		[format appendFormat:@"%d min ", min];
	if (sec!=0)
		[format appendFormat:@"%d sec", sec];
	

	
	return format;
		
}

- (void)interrupt
{
	interrupt = YES;
}

@end
